import math
import mock
import os
import pkg_resources
import platform
import pytest
import time
import unittest2
from uuid import uuid4

from kazoo.client import KazooClient
from kazoo.handlers.gevent import SequentialGeventHandler

from pykafka import KafkaClient
from pykafka.balancedconsumer import BalancedConsumer, OffsetType
from pykafka.exceptions import ConsumerStoppedException
from pykafka.managedbalancedconsumer import ManagedBalancedConsumer
from pykafka.test.utils import get_cluster, stop_cluster
from pykafka.utils.compat import range, iterkeys, iteritems
from tests.pykafka import patch_subclass


kafka_version_string = os.environ.get('KAFKA_VERSION', '0.8')
kafka_version = pkg_resources.parse_version(kafka_version_string)
version_09 = pkg_resources.parse_version("0.9.0.0")


class TestBalancedConsumer(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._consumer_timeout = 2000
        cls._mock_consumer, _ = TestBalancedConsumer.buildMockConsumer(timeout=cls._consumer_timeout)

    @classmethod
    def buildMockConsumer(self, num_partitions=10, num_participants=1, timeout=2000):
        consumer_group = b'testgroup'
        topic = mock.Mock()
        topic.name = 'testtopic'
        topic.partitions = {}
        for k in range(num_partitions):
            part = mock.Mock(name='part-{part}'.format(part=k))
            part.id = k
            part.topic = topic
            part.leader = mock.Mock()
            part.leader.id = k % num_participants
            topic.partitions[k] = part

        cluster = mock.MagicMock()
        zk = mock.MagicMock()
        return BalancedConsumer(topic, cluster, consumer_group,
                                zookeeper=zk, auto_start=False, use_rdkafka=False,
                                consumer_timeout_ms=timeout), topic

    def test_consume_returns(self):
        """Ensure that consume() returns in the amount of time it's supposed to
        """
        self._mock_consumer._setup_internal_consumer(start=False)
        self._mock_consumer._consumer._partitions_by_id = {1: "dummy"}
        self._mock_consumer._running = True
        start = time.time()
        self._mock_consumer.consume()
        self.assertEqual(int(time.time() - start), int(self._consumer_timeout / 1000))

    def test_consume_graceful_stop(self):
        """Ensure that stopping a consumer while consuming from Kafka does not
        end in an infinite loop when timeout is not used.
        """
        consumer, _ = self.buildMockConsumer(timeout=-1)
        consumer._setup_internal_consumer(start=False)
        consumer._consumer._partitions_by_id = {1: "dummy"}

        consumer.stop()
        with self.assertRaises(ConsumerStoppedException):
            consumer.consume()

    def test_decide_partitions(self):
        """Test partition assignment for a number of partitions/consumers."""
        # 100 test iterations
        for i in range(100):
            # Set up partitions, cluster, etc
            num_participants = i + 1
            num_partitions = 100 - i
            participants = sorted(['test-debian:{p}'.format(p=p)
                                   for p in range(num_participants)])
            cns, topic = self.buildMockConsumer(num_partitions=num_partitions,
                                                num_participants=num_participants)

            # Simulate each participant to ensure they're correct
            assigned_parts = []
            for p_id in range(num_participants):
                cns._consumer_id = participants[p_id]  # override consumer id

                # Decide partitions then validate
                partitions = cns._decide_partitions(participants)
                assigned_parts.extend(partitions)

                remainder_ppc = num_partitions % num_participants
                idx = participants.index(cns._consumer_id)
                parts_per_consumer = num_partitions / num_participants
                parts_per_consumer = math.floor(parts_per_consumer)

                num_parts = parts_per_consumer + (0 if (idx + 1 > remainder_ppc) else 1)

                self.assertEqual(len(partitions), int(num_parts))

            # Validate all partitions were assigned once and only once
            all_partitions = topic.partitions.values()
            all_partitions = sorted(all_partitions, key=lambda x: x.id)
            assigned_parts = sorted(assigned_parts, key=lambda x: x.id)
            self.assertListEqual(assigned_parts, all_partitions)


class TestManagedBalancedConsumer(TestBalancedConsumer):
    @classmethod
    def buildMockConsumer(self, num_partitions=10, num_participants=1, timeout=2000):
        consumer_group = b'testgroup'
        topic = mock.Mock()
        topic.name = 'testtopic'
        topic.partitions = {}
        for k in range(num_partitions):
            part = mock.Mock(name='part-{part}'.format(part=k))
            part.id = k
            part.topic = topic
            part.leader = mock.Mock()
            part.leader.id = k % num_participants
            topic.partitions[k] = part

        cluster = mock.MagicMock()
        cns = ManagedBalancedConsumer(topic, cluster, consumer_group,
                                      auto_start=False, use_rdkafka=False,
                                      consumer_timeout_ms=timeout)
        cns._group_coordinator = mock.MagicMock()
        return cns, topic


class BalancedConsumerIntegrationTests(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False
    MANAGED_CONSUMER = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = uuid4().hex.encode()
        cls.n_partitions = 3
        cls.kafka.create_topic(cls.topic_name, cls.n_partitions, 2)
        cls.client = KafkaClient(cls.kafka.brokers,
                                 use_greenlets=cls.USE_GEVENT,
                                 broker_version=kafka_version_string)
        cls.prod = cls.client.topics[cls.topic_name].get_producer(
            min_queued_messages=1
        )
        for i in range(1000):
            cls.prod.produce('msg {num}'.format(num=i).encode())

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def get_zk(self):
        if not self.USE_GEVENT:
            return KazooClient(self.kafka.zookeeper)
        return KazooClient(self.kafka.zookeeper, handler=SequentialGeventHandler())

    def get_balanced_consumer(self, consumer_group, **kwargs):
        if self.MANAGED_CONSUMER:
            kwargs.pop("zookeeper", None)
            kwargs.pop("zookeeper_connect", None)
        return self.client.topics[self.topic_name].get_balanced_consumer(
            consumer_group,
            managed=self.MANAGED_CONSUMER,
            **kwargs
        )

    def test_extra_consumer(self):
        """Ensure proper operation of "extra" consumers in a group

        An "extra" consumer is the N+1th member of a consumer group consuming a topic
        of N partitions, and any consumer beyond the N+1th.
        """
        group = b"test_extra_consumer"
        extras = 1

        def verify_extras(consumers, extras_count):
            messages = [c.consume() for c in consumers]
            successes = [a for a in messages if a is not None]
            nones = [a for a in messages if a is None]
            attempts = 0
            while len(nones) != extras_count and attempts < 5:
                messages = [c.consume() for c in consumers]
                successes = [a for a in messages if a is not None]
                nones = [a for a in messages if a is None]
                attempts += 1
            self.assertEqual(len(nones), extras_count)
            self.assertEqual(len(successes), self.n_partitions)

        try:
            consumers = [self.get_balanced_consumer(group, consumer_timeout_ms=5000)
                         for i in range(self.n_partitions + extras)]
            verify_extras(consumers, extras)

            # when one consumer stops, the extra should pick up its partitions
            removed = consumers[:extras]
            for consumer in removed:
                consumer.stop()
            consumers = [a for a in consumers if a not in removed]
            self.wait_for_rebalancing(*consumers)
            self.assertEqual(len(consumers), self.n_partitions)
            verify_extras(consumers, 0)

            # added "extra" consumers should idle
            for i in range(extras):
                consumers.append(self.get_balanced_consumer(group,
                                                            consumer_timeout_ms=5000))
            self.wait_for_rebalancing(*consumers)
            verify_extras(consumers, extras)
        finally:
            for consumer in consumers:
                try:
                    consumer.stop()
                except:
                    pass

    def test_rebalance_callbacks(self):
        def on_rebalance(cns, old_partition_offsets, new_partition_offsets):
            self.assertTrue(len(new_partition_offsets) > 0)
            self.assigned_called = True
            for id_ in iterkeys(new_partition_offsets):
                new_partition_offsets[id_] = self.offset_reset
            return new_partition_offsets

        self.assigned_called = False
        self.offset_reset = 50
        try:
            consumer_group = b'test_rebalance_callbacks'
            consumer_a = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                post_rebalance_callback=on_rebalance,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)
            self.wait_for_rebalancing(consumer_a, consumer_b)
            self.assertTrue(self.assigned_called)
            for _, offset in iteritems(consumer_a.held_offsets):
                self.assertEqual(offset, self.offset_reset)
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_rebalance_callbacks_surfaces_errors(self):
        def on_rebalance(cns, old_partition_offsets, new_partition_offsets):
            raise ValueError("BAD CALLBACK")

        self.assigned_called = False
        self.offset_reset = 50
        try:
            consumer_group = b'test_rebalance_callbacks_error'
            consumer_a = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                post_rebalance_callback=on_rebalance,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)

            with pytest.raises(ValueError) as ex:
                self.wait_for_rebalancing(consumer_a, consumer_b)
                assert 'BAD CALLBACK' in str(ex.value)

        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_consume_earliest(self):
        try:
            consumer_a = self.get_balanced_consumer(
                b'test_consume_earliest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                b'test_consume_earliest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)

            # Consume from both a few times
            messages = [consumer_a.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)
            messages = [consumer_b.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)

            # Validate they aren't sharing partitions
            self.assertSetEqual(
                consumer_a._partitions & consumer_b._partitions,
                set()
            )

            # Validate all partitions are here
            self.assertSetEqual(
                consumer_a._partitions | consumer_b._partitions,
                set(self.client.topics[self.topic_name].partitions.values())
            )
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_consume_latest(self):
        try:
            consumer_a = self.get_balanced_consumer(
                b'test_consume_latest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.LATEST,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                b'test_consume_latest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.LATEST,
                use_rdkafka=self.USE_RDKAFKA)

            # Make sure we're done before producing more messages:
            self.wait_for_rebalancing(consumer_a, consumer_b)

            # Since we are consuming from the latest offset,
            # produce more messages to consume.
            for i in range(10):
                self.prod.produce('msg {num}'.format(num=i).encode())

            # Consume from both a few times
            messages = [consumer_a.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)
            messages = [consumer_b.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)

            # Validate they aren't sharing partitions
            self.assertSetEqual(
                consumer_a._partitions & consumer_b._partitions,
                set()
            )

            # Validate all partitions are here
            self.assertSetEqual(
                consumer_a._partitions | consumer_b._partitions,
                set(self.client.topics[self.topic_name].partitions.values())
            )
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_external_kazoo_client(self):
        """Run with pre-existing KazooClient instance

        This currently doesn't assert anything, it just rules out any trivial
        exceptions in the code path that uses an external KazooClient
        """
        zk = KazooClient(self.kafka.zookeeper)
        zk.start()

        consumer = self.get_balanced_consumer(
            b'test_external_kazoo_client',
            zookeeper=zk,
            consumer_timeout_ms=10,
            use_rdkafka=self.USE_RDKAFKA)
        [msg for msg in consumer]
        consumer.stop()
    test_external_kazoo_client.skip_condition = lambda cls: cls.MANAGED_CONSUMER

    def test_no_partitions(self):
        """Ensure a consumer assigned no partitions doesn't fail"""

        def _decide_dummy(p, consumer_id=None):
            return set()
        consumer = self.get_balanced_consumer(
            b'test_no_partitions',
            zookeeper_connect=self.kafka.zookeeper,
            auto_start=False,
            consumer_timeout_ms=50,
            use_rdkafka=self.USE_RDKAFKA)

        consumer._decide_partitions = _decide_dummy
        consumer.start()
        res = consumer.consume()
        self.assertEqual(res, None)
        self.assertTrue(consumer._running)
        # check that stop() succeeds (cf #313 and #392)
        consumer.stop()

    def test_zk_conn_lost(self):
        """Check we restore zookeeper nodes correctly after connection loss

        See also github issue #204.
        """
        check_partitions = lambda c: c._get_held_partitions() == c._partitions
        zk = self.get_zk()
        zk.start()
        try:
            consumer_group = b'test_zk_conn_lost'

            consumer = self.get_balanced_consumer(consumer_group,
                                                  zookeeper=zk,
                                                  use_rdkafka=self.USE_RDKAFKA)
            self.assertTrue(check_partitions(consumer))
            with consumer._rebalancing_lock:
                zk.stop()  # expires session, dropping all our nodes

            # Start a second consumer on a different zk connection
            other_consumer = self.get_balanced_consumer(
                consumer_group, use_rdkafka=self.USE_RDKAFKA)

            # Slightly contrived: we'll grab a lock to keep _rebalance() from
            # starting when we restart the zk connection (restart triggers a
            # rebalance), so we can confirm the expected discrepancy between
            # the (empty) set of partitions on zk and the set in the internal
            # consumer:
            with consumer._rebalancing_lock:
                zk.start()
                self.assertFalse(check_partitions(consumer))

            # Finally, confirm that _rebalance() resolves the discrepancy:
            self.wait_for_rebalancing(consumer, other_consumer)
            self.assertTrue(check_partitions(consumer))
            self.assertTrue(check_partitions(other_consumer))
        finally:
            try:
                consumer.stop()
                other_consumer.stop()
                zk.stop()
            except:
                pass
    test_zk_conn_lost.skip_condition = lambda cls: cls.MANAGED_CONSUMER

    def wait_for_rebalancing(self, *balanced_consumers):
        """Test helper that loops while rebalancing is ongoing

        Needs to be given all consumer instances active in a consumer group.
        Waits for up to 100 seconds, which should be enough for even a very
        oversubscribed test cluster.
        """
        for _ in range(500):
            n_parts = [len(cons.partitions) for cons in balanced_consumers]
            if (max(n_parts) - min(n_parts) <= 1
                    and sum(n_parts) == self.n_partitions):
                break
            else:
                balanced_consumers[0]._cluster.handler.sleep(.2)
            # check for failed consumers (there'd be no point waiting anymore)
            [cons._raise_worker_exceptions() for cons in balanced_consumers]
        else:
            raise AssertionError("Rebalancing failed")


@patch_subclass(BalancedConsumerIntegrationTests,
                platform.python_implementation() == "PyPy")
class BalancedConsumerGEventIntegrationTests(unittest2.TestCase):
    USE_GEVENT = True


@patch_subclass(BalancedConsumerIntegrationTests, kafka_version < version_09)
class ManagedBalancedConsumerIntegrationTests(unittest2.TestCase):
    MANAGED_CONSUMER = True


@patch_subclass(
    BalancedConsumerIntegrationTests,
    platform.python_implementation() == "PyPy" or kafka_version < version_09)
class ManagedBalancedConsumerGEventIntegrationTests(unittest2.TestCase):
    MANAGED_CONSUMER = True
    USE_GEVENT = True


if __name__ == "__main__":
    unittest2.main()



from contextlib import contextmanager
import json
import mock
import os
import platform
import pytest
import time
import unittest2
from uuid import uuid4

from pykafka import KafkaClient
from pykafka.simpleconsumer import OwnedPartition, OffsetType
from pykafka.test.utils import get_cluster, stop_cluster
from pykafka.utils.compat import range, iteritems, get_string


kafka_version = os.environ.get('KAFKA_VERSION', '0.8.0')


class TestSimpleConsumer(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = uuid4().hex.encode()
        cls.kafka.create_topic(cls.topic_name, 3, 2)

        cls.total_msgs = 1000
        cls.client = KafkaClient(cls.kafka.brokers, broker_version=kafka_version)
        cls.prod = cls.client.topics[cls.topic_name].get_producer(
            min_queued_messages=1
        )
        for i in range(cls.total_msgs):
            cls.prod.produce('msg {i}'.format(i=i).encode())

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    @contextmanager
    def _get_simple_consumer(self, **kwargs):
        topic = self.client.topics[self.topic_name]
        consumer = topic.get_simple_consumer(
            use_rdkafka=self.USE_RDKAFKA, **kwargs)
        try:
            yield consumer
        finally:
            consumer.stop()

    def test_consume(self):
        """Test consuming all messages in topic"""
        # This uses a fairly long timeout to allow the test to pass on an
        # oversubscribed test cluster
        with self._get_simple_consumer(consumer_timeout_ms=30000) as consumer:
            count = 0
            for msg in consumer:
                self.assertIsNotNone(msg.value)
                count += 1
                if count == self.total_msgs:
                    # We don't want to wait for StopIteration, given the long
                    # timeout set above
                    break
            self.assertEquals(count, self.total_msgs)

    @staticmethod
    def _convert_offsets(offset_responses):
        """Helper function to translate Offset(Fetch)PartitionResponse

        Calls like consumer.fetch_offsets() and earliest_available_offsets()
        return lists of OffsetPartitionResponses.  These hold the next offset
        to be consumed, whereas consumer.held_offsets returns the latest
        consumed offset.  This translates them to facilitate comparisons.
        """
        if isinstance(offset_responses, dict):
            offset_responses = iteritems(offset_responses)
        f1 = lambda off: OffsetType.EARLIEST if off == 0 else off - 1
        f2 = lambda off: off[0] if isinstance(off, list) else off
        return {partition_id: f1(f2(offset_response.offset))
                for partition_id, offset_response in offset_responses}

    def test_offset_commit(self):
        """Check fetched offsets match pre-commit internal state"""
        with self._get_simple_consumer(
                consumer_group=b'test_offset_commit') as consumer:
            [consumer.consume() for _ in range(100)]
            offsets_committed = consumer.held_offsets
            consumer.commit_offsets()

            offsets_fetched = self._convert_offsets(consumer.fetch_offsets())
            self.assertEquals(offsets_fetched, offsets_committed)

    def test_offset_resume(self):
        """Check resumed internal state matches committed offsets"""
        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume') as consumer:
            [consumer.consume() for _ in range(100)]
            offsets_committed = consumer.held_offsets
            consumer.commit_offsets()

        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume') as consumer:
            self.assertEquals(consumer.held_offsets, offsets_committed)

    def test_reset_offset_on_start(self):
        """Try starting from LATEST and EARLIEST offsets"""
        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.EARLIEST,
                reset_offset_on_start=True) as consumer:
            earliest_offs = self._convert_offsets(
                consumer.topic.earliest_available_offsets())
            self.assertEquals(earliest_offs, consumer.held_offsets)
            self.assertIsNotNone(consumer.consume())

        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.LATEST,
                reset_offset_on_start=True,
                consumer_timeout_ms=500) as consumer:
            latest_offs = self._convert_offsets(
                consumer.topic.latest_available_offsets())
            self.assertEquals(latest_offs, consumer.held_offsets)
            self.assertIsNone(consumer.consume(block=False))

        difference = sum(latest_offs[i] - earliest_offs[i]
                         if earliest_offs[i] >= 0 else latest_offs[i] + 1
                         if latest_offs[i] >= 0 else 0
                         for i in latest_offs)
        self.assertEqual(difference, self.total_msgs)

    def test_reset_offsets(self):
        """Test resetting to user-provided offsets"""
        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.EARLIEST) as consumer:
            # Find us a non-empty partition "target_part"
            part_id, latest_offset = next(
                (p, res.offset[0])
                for p, res in consumer.topic.latest_available_offsets().items()
                if res.offset[0] > 0)
            target_part = consumer.partitions[part_id]

            # Set all other partitions to LATEST, to ensure that any consume()
            # calls read from target_part
            partition_offsets = {
                p: OffsetType.LATEST for p in consumer.partitions.values()}

            new_offset = latest_offset - 5
            partition_offsets[target_part] = new_offset
            consumer.reset_offsets(partition_offsets.items())

            self.assertEqual(consumer.held_offsets[part_id], new_offset)
            msg = consumer.consume()
            self.assertEqual(msg.offset, new_offset + 1)

            # Invalid offsets should get overwritten as per auto_offset_reset
            partition_offsets[target_part] = latest_offset + 5  # invalid!
            consumer.reset_offsets(partition_offsets.items())

            # SimpleConsumer's fetcher thread will detect the invalid offset
            # and reset it immediately.  RdKafkaSimpleConsumer however will
            # only get to write the valid offset upon a call to consume():
            msg = consumer.consume()
            expected_offset = target_part.earliest_available_offset()
            self.assertEqual(msg.offset, expected_offset)
            self.assertEqual(consumer.held_offsets[part_id], expected_offset)

    def test_update_cluster(self):
        """Check that the consumer can initiate cluster updates"""
        with self._get_simple_consumer() as consumer:
            self.assertIsNotNone(consumer.consume())

            for broker in self.client.brokers.values():
                broker._connection.disconnect()

            # The consumer fetcher thread should prompt broker reconnection
            t_start = time.time()
            timeout = 10.
            try:
                for broker in self.client.brokers.values():
                    while not broker._connection.connected:
                        time.sleep(.1)
                        self.assertTrue(time.time() - t_start < timeout,
                                        msg="Broker reconnect failed.")
            finally:
                # Make sure further tests don't get confused
                consumer._update()
            # If the fetcher thread fell over during the cluster update
            # process, we'd get an exception here:
            self.assertIsNotNone(consumer.consume())

    def test_consumer_lag(self):
        """Ensure that after consuming the entire topic, lag is 0"""
        with self._get_simple_consumer(consumer_group=b"test_lag_group",
                                       consumer_timeout_ms=1000) as consumer:
            while True:
                message = consumer.consume()
                if message is None:
                    break
            consumer.commit_offsets()
            latest_offsets = {p_id: res.offset[0]
                              for p_id, res
                              in iteritems(consumer.topic.latest_available_offsets())}
            current_offsets = {p_id: res.offset for p_id, res in consumer.fetch_offsets()}
            self.assertEqual(current_offsets, latest_offsets)


@pytest.mark.skipif(platform.python_implementation() == "PyPy",
                    reason="Unresolved crashes")
class TestGEventSimpleConsumer(TestSimpleConsumer):
    USE_GEVENT = True


class TestOwnedPartition(unittest2.TestCase):
    def test_partition_saves_offset(self):
        offset = 20
        msgval = "test"
        partition = mock.MagicMock()
        op = OwnedPartition(partition)
        op.next_offset = offset

        message = mock.Mock()
        message.value = msgval
        message.offset = offset

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 1)
        ret_message = op.consume()
        self.assertEqual(op.last_offset_consumed, message.offset)
        self.assertEqual(op.next_offset, message.offset + 1)
        self.assertNotEqual(ret_message, None)
        self.assertEqual(ret_message.value, msgval)

    def test_partition_rejects_old_message(self):
        last_offset = 400
        op = OwnedPartition(None)
        op.last_offset_consumed = last_offset

        message = mock.Mock()
        message.value = "test"
        message.offset = 20

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 0)
        op.consume()
        self.assertEqual(op.last_offset_consumed, last_offset)

    def test_compacted_topic_partition_rejects_old_message_after_initial(self):
        last_offset = 400
        message1 = mock.Mock()
        message1.value = "first-test"
        message1.partition_id = 0
        message1.offset = last_offset

        partition = mock.MagicMock()
        partition.id = 0
        op = OwnedPartition(partition, compacted_topic=True)
        op.enqueue_messages([message1])
        self.assertEqual(op.message_count, 1)
        consumed_msg = op.consume()
        self.assertEqual(op.message_count, 0)
        self.assertEqual(op.last_offset_consumed, last_offset)

        message2 = mock.Mock()
        message2.value = "test"
        message2.partition_id = 0
        message2.offset = 20

        op.enqueue_messages([message2])
        self.assertEqual(op.message_count, 0)
        op.consume()
        self.assertEqual(op.last_offset_consumed, last_offset)

    def test_partition_consume_empty_queue(self):
        op = OwnedPartition(None)

        message = op.consume()
        self.assertEqual(message, None)

    def test_partition_offset_commit_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)
        op.last_offset_consumed = 200

        request = op.build_offset_commit_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)
        self.assertEqual(request.offset, op.last_offset_consumed + 1)
        parsed_metadata = json.loads(get_string(request.metadata))
        self.assertEqual(parsed_metadata["consumer_id"], '')
        self.assertTrue(bool(parsed_metadata["hostname"]))

    def test_partition_offset_fetch_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)

        request = op.build_offset_fetch_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)

    def test_partition_offset_counters(self):
        res = mock.Mock()
        res.offset = 400

        op = OwnedPartition(None)
        op.set_offset(res.offset)

        self.assertEqual(op.last_offset_consumed, res.offset)
        self.assertEqual(op.next_offset, res.offset + 1)


if __name__ == "__main__":
    unittest2.main()



import operator
import unittest2

from pykafka import protocol
from pykafka.common import CompressionType
from pykafka.utils.compat import buffer


class TestMetadataAPI(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequest()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x15'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x00'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id
                # end header

                b'\x00\x00\x00\x00'  # len(topics)
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node is
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x03'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(response.topics[b'test'].partitions[0].err, 3)

    def test_topic_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x03'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(response.topics[b'test'].err, 3)


class TestProduceAPI(unittest2.TestCase):
    maxDiff = None

    test_messages = [
        protocol.Message(b'this is a test message', partition_key=b'asdf'),
        protocol.Message(b'this is also a test message', partition_key=b'test_key'),
        protocol.Message(b"this doesn't have a partition key"),
    ]

    def test_request(self):
        message = self.test_messages[0]
        req = protocol.ProduceRequest()
        req.add_message(message, b'test', 0)
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b"\x00\x00\x00a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07pykafka"  # header
                b'\x00\x01'  # required acks
                b'\x00\x00\'\x10'  # timeout
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len (partitions)
                        b'\x00\x00\x00\x00'  # partition
                        b'\x00\x00\x004'  # message set size
                            b'\xff\xff\xff\xff\xff\xff\xff\xff'  # offset
                            b'\x00\x00\x00('  # message size
                                b'\x0e\x8a\x19O'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x04'  # len(key)
                                    b'asdf'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b"this is a test message"  # value
            )
        )

    def test_gzip_compression(self):
        req = protocol.ProduceRequest(compression_type=CompressionType.GZIP)
        [req.add_message(m, b'test_gzip', 0) for m in self.test_messages]
        msg = req.get_bytes()
        self.assertEqual(len(msg), 207)  # this isn't a good test

    def test_snappy_compression(self):
        req = protocol.ProduceRequest(compression_type=CompressionType.SNAPPY)
        [req.add_message(m, b'test_snappy', 0) for m in self.test_messages]
        msg = req.get_bytes()
        self.assertEqual(len(msg), 212)  # this isn't a good test

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.ProduceResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x03'  # error code
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        response = protocol.ProduceResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00'  # error code
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(
            response.topics,
            {b'test': {0: protocol.ProducePartitionResponse(0, 2)}}
        )


class TestFetchAPI(unittest2.TestCase):
    maxDiff = None

    expected_data = [
        {
            'partition_key': b'asdf',
            'compression_type': 0,
            'value': b'this is a test message',
            'offset': 0,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'partition': None
        }, {
            'partition_key': b'test_key',
            'compression_type': 0,
            'value': b'this is also a test message',
            'offset': 1,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'partition': None
        }, {
            'partition_key': None,
            'compression_type': 0,
            'value': b"this doesn't have a partition key",
            'offset': 2,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'partition': None
        }]

    def msg_to_dict(self, msg):
        """Helper to extract data from Message slots"""
        attr_names = protocol.Message.__slots__
        f = operator.attrgetter(*attr_names)
        return dict(zip(attr_names, f(msg)))

    def test_request(self):
        preq = protocol.PartitionFetchRequest(b'test', 0, 1)
        req = protocol.FetchRequest(partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00;\x00\x01\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\xff\xff\xff\xff'  # replica id
                b'\x00\x00\x03\xe8'  # max wait time
                b'\x00\x00\x04\x00'  # min bytes
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # fetch offset
                            b'\x00\x10\x00\x00'  # max bytes
            )
        )

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.FetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x03'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = protocol.FetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x00'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(len(resp.topics[b'test'][0].messages), 1)
        self.assertEqual(resp.topics[b'test'][0].max_offset, 2)
        message = resp.topics[b'test'][0].messages[0]
        self.assertEqual(message.value, b'this is a test message')
        self.assertEqual(message.partition_key, b'test_partition_key')
        self.assertEqual(message.compression_type, 0)
        self.assertEqual(message.offset, 1)

    def test_gzip_decompression(self):
        msg = b''.join([
            b'\x00\x00\x00\x01'  # len(topics)
            b'\x00\t'  # len(topic name)
                b'test_gzip'  # topic name
            b'\x00\x00\x00\x01'  # len(partitions)
                b'\x00\x00\x00\x00'  # partition
                    b'\x00\x00'  # error code
                    b'\x00\x00\x00\x00\x00\x00\x00\x03'  # highwater mark offset
                    b'\x00\x00\x00\x9b'  # message set size
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
                        b'\x00\x00\x00\x8f'  # message size
                            b'\xbb\xe7\x1f\xb8'  # crc
                            b'\x00'  # magic byte
                            b'\x01'  # attributes
                            b'\xff\xff\xff\xff'  # len(key)
                            b'\x00\x00\x00\x81'  # len(value)
                                b'\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x03\r\xbe.I\x7f0\x8b%\xb18%\rH\x8b\x95dd\x16+\x00Q\xa2BIjq\x89Bnjqqbz*T=#\x10\x1b\xb2\xf3\xcb\xf4\x81y\x1c \x15\xf1\xd9\xa9\x95@\xb64\\_Nq>v\xcdL@\xac\x7f\xb5(\xd9\x98\x81\xe1?\x10\x00y\x8a`M)\xf9\xa9\xc5y\xea%\n\x19\x89e\xa9@\x9d\x05\x89E%\x99%\x99\xf9y\n@\x93\x01N1\x9f[\xac\x00\x00\x00'  # value
        ])
        response = protocol.FetchResponse(msg)
        for i in range(len(self.expected_data)):
            self.assertDictEqual(
                self.msg_to_dict(response.topics[b'test_gzip'][0].messages[i]),
                self.expected_data[i])

    def test_snappy_decompression(self):
        msg = b''.join([
            b'\x00\x00\x00\x01'  # len(topics)
            b'\x00\x0b'  # len(topic name)
                b'test_snappy'  # topic name
            b'\x00\x00\x00\x01'  # len(partitions)
                b'\x00\x00\x00\x00'  # partition
                    b'\x00\x00'  # error code
                    b'\x00\x00\x00\x00\x00\x00\x00\x03'  # highwater mark offset
                    b'\x00\x00\x00\xb5'  # message set size
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
                        b'\x00\x00\x00\xa9'  # message size
                            b'\xc1\xf2\xa3\xe1'  # crc
                            b'\x00'  # magic bytes
                            b'\x02'  # attributes
                            b'\xff\xff\xff\xff'  # len(key)
                            b'\x00\x00\x00\x9b'  # len(value)
                                b'\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x87\xac\x01\x00\x00\x19\x01\x10(\x0e\x8a\x19O\x05\x0fx\x04asdf\x00\x00\x00\x16this is a test message\x05$(\x00\x00\x01\x00\x00\x001\x07\x0f\x1c\x8e\x05\x10\x00\x08\x01"\x1c_key\x00\x00\x00\x1b\x158\x08lsoV=\x00H\x02\x00\x00\x00/\xd5rc3\x00\x00\xff\xff\xff\xff\x00\x00\x00!\x055ldoesn\'t have a partition key'  # value
        ])
        response = protocol.FetchResponse(msg)
        for i in range(len(self.expected_data)):
            self.assertDictEqual(
                self.msg_to_dict(response.topics[b'test_snappy'][0].messages[i]),
                self.expected_data[i])


class TestOffsetAPI(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        preq = protocol.PartitionOffsetRequest(b'test', 0, -1, 1)
        req = protocol.OffsetRequest(partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x003\x00\x02\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\xff\xff\xff\xff'  # replica id
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                        b'\xff\xff\xff\xff\xff\xff\xff\xff'  # time
                        b'\x00\x00\x00\x01'  # max number of offsets
            )
        )

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.OffsetResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x03'  # error code
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = protocol.OffsetResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x00'  # error code
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(resp.topics[b'test'][0].offset, [2])


class TestOffsetCommitFetchAPI(unittest2.TestCase):
    maxDiff = None

    def test_consumer_metadata_request(self):
        req = protocol.GroupCoordinatorRequest(b'test')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x17\x00\n\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x04'  # len(group id)
                    b'test'  # group id
            )
        )

    def test_consumer_metadata_response(self):
        response = protocol.GroupCoordinatorResponse(
            buffer(
                b'\x00\x00'  # error code
                b'\x00\x00\x00\x00'  # coordinator id
                b'\x00\r'  # len(coordinator host)
                    b'emmett-debian'  # coordinator host
                b'\x00\x00#\x84'  # coordinator port
            )
        )
        self.assertEqual(response.coordinator_id, 0)
        self.assertEqual(response.coordinator_host, b'emmett-debian')
        self.assertEqual(response.coordinator_port, 9092)

    def test_offset_commit_request(self):
        preq = protocol.PartitionOffsetCommitRequest(
            b'test', 0, 68, 1426632066, b'testmetadata')
        req = protocol.OffsetCommitRequest(
            b'test', 1, b'pykafka', partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00T\x00\x08\x00\x01\x00\x00\x00\x00\x00\x07pykafka'
                b'\x00\x04'  # len(consumer group id)
                    b'test'  # consumer group id
                b'\x00\x00\x00\x01'  # consumer group generation id
                b'\x00\x07'  # len(consumer id)
                    b'pykafka'  # consumer id
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                b'\x00\x00\x00\x01'  # len(partitions)
                    b'\x00\x00\x00\x00'  # partition
                        b'\x00\x00\x00\x00\x00\x00\x00D'  # offset
                        b'\x00\x00\x00\x00U\x08\xad\x82'  # timestamp
                        b'\x00\x0c'  # len(metadata)
                            b'testmetadata'  # metadata
            )
        )

    def test_offset_commit_response(self):
        response = protocol.OffsetCommitResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x0c'  # len(topic name)
                        b'emmett.dummy'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.topics[b'emmett.dummy'][0].err, 0)

    def test_offset_fetch_request(self):
        preq = protocol.PartitionOffsetFetchRequest(b'testtopic', 0)
        req = protocol.OffsetFetchRequest(b'test', partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00.\x00\t\x00\x01\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x04'  # len(consumer group)
                    b'test'  # consumer group
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\t'  # len(topic name)
                        b'testtopic'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
            )
        )

    def test_offset_fetch_response(self):
        response = protocol.OffsetFetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x0c'  # len(topic name)
                        b'emmett.dummy'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00'  # len(metadata)
                            b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.topics[b'emmett.dummy'][0].metadata, b'')
        self.assertEqual(response.topics[b'emmett.dummy'][0].offset, 1)


class TestGroupMembershipAPI(unittest2.TestCase):
    maxDiff = None

    def test_consumer_group_protocol_metadata(self):
        meta = protocol.ConsumerGroupProtocolMetadata()
        msg = meta.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x00'  # version
                b'\x00\x01'  # len(subscription)
                    b'\x00\n'  # len(topic name)
                        b'dummytopic'  # topic name
                    b'\x00\x00\x00\x0c'  # len(userdata)
                        b'testuserdata')  # userdata
        )

    def test_join_group_request(self):
        req = protocol.JoinGroupRequest(b'dummygroup', member_id=b'testmember')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00|\x00\x0b\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(groupid)
                    b'dummygroup'  # groupid
                b'\x00\x00u0'  # session timeout
                b'\x00\n'  # len(memberid)
                    b'testmember'  # memberid
                b'\x00\x08'  # len(protocol type)
                    b'consumer'  # protocol type
                b'\x00\x00\x00\x01'  # len(group protocols)
                    b'\x00\x19'  # len(protocol name)
                        b'pykafkaassignmentstrategy'  # protocol name
                    b'\x00\x00\x00"'  # len(protocol metadata)
                        b'\x00\x00\x00\x00\x00\x01\x00\ndummytopic\x00\x00\x00\x0ctestuserdata'  # protocol metadata
            )
        )

    def test_join_group_response(self):
        response = protocol.JoinGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\x17'  # len (group protocol)
                    b'dummyassignmentstrategy'  # group protocol
                b'\x00,'  # len(leader id)
                    b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # leader id
                b'\x00,'  # len(member id)
                    b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # member id
                b'\x00\x00\x00\x01'  # leb(members)
                    b'\x00,'  # len(member id)
                        b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # member id
                    b'\x00\x00\x00"'  # len(member metadata)
                        b'\x00\x00\x00\x00\x00\x01\x00\ndummytopic\x00\x00\x00\x0ctestuserdata\x00\x00\x00\x00'  # member metadata
            )
        )
        self.assertEqual(response.generation_id, 1)
        self.assertEqual(response.group_protocol, b'dummyassignmentstrategy')
        self.assertEqual(response.leader_id,
                         b'pykafka-b2361322-674c-4e26-9194-305962636e57')
        self.assertEqual(response.member_id,
                         b'pykafka-b2361322-674c-4e26-9194-305962636e57')
        self.assertEqual(response.members,
                         {b'pykafka-b2361322-674c-4e26-9194-305962636e57': b'\x00\x00\x00\x00\x00\x01\x00\ndummytopic\x00\x00\x00\x0ctestuserdata'})

    def test_member_assignment_construction(self):
        assignment = protocol.MemberAssignment([(b"mytopic1", [3, 5, 7, 9]),
                                                (b"mytopic2", [2, 4, 6, 8])])
        msg = assignment.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x01'  # version
                b'\x00\x00\x00\x02'  # len(partition assignment)
                    b'\x00\x08'  # len(topic)
                        b'mytopic1'  # topic
                    b'\x00\x00\x00\x04'  # len(partitions)
                        b'\x00\x00\x00\x03'  # partition
                        b'\x00\x00\x00\x05'  # partition
                        b'\x00\x00\x00\x07'  # partition
                        b'\x00\x00\x00\t'  # partition
                    b'\x00\x08'  # len(topic)
                        b'mytopic2'  # topic
                    b'\x00\x00\x00\x04'  # len(partitions)
                        b'\x00\x00\x00\x02'  # partition
                        b'\x00\x00\x00\x04'  # partition
                        b'\x00\x00\x00\x06'  # partition
                        b'\x00\x00\x00\x08'  # partition
            )
        )

    def test_sync_group_request(self):
        req = protocol.SyncGroupRequest(
            b'dummygroup', 1, b'testmember1',
            [
                protocol.MemberAssignment([(b"mytopic1", [3, 5, 7, 9]),
                                           (b"mytopic2", [3, 5, 7, 9])], member_id=b"a"),
                protocol.MemberAssignment([(b"mytopic1", [2, 4, 6, 8]),
                                           (b"mytopic2", [2, 4, 6, 8])], member_id=b"b")
            ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\xc4\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\x0b'  # len(member id)
                    b'testmember1'  # member id
                b'\x00\x00\x00\x02'  # len(group assignment)
                    b'\x00\x01'  # len(member id)
                        b'a'  # member id
                    b'\x00\x00\x00B'  # len(member assignment)
                        b'\x00\x01\x00\x00\x00\x02\x00\x08mytopic1\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x07\x00\x00\x00\t\x00\x08mytopic2\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x07\x00\x00\x00\t'  # member assignment
                    b'\x00\x01'  # len(member id)
                        b'b'  # member id
                    b'\x00\x00\x00B'  # len(member assignment)
                        b'\x00\x01\x00\x00\x00\x02\x00\x08mytopic1\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x06\x00\x00\x00\x08\x00\x08mytopic2\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x06\x00\x00\x00\x08'  # member assignment
            )
        )

    def test_sync_group_response(self):
        response = protocol.SyncGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
                b'\x00\x00\x00H'  # len(member assignment)
                    b'\x00\x01\x00\x00\x00\x01\x00\x14testtopic_replicated\x00\x00\x00\n\x00\x00\x00\x06\x00\x00\x00\x07\x00\x00\x00\x08\x00\x00\x00\t\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05,pyk'  # member assignment
            )
        )
        self.assertEqual(response.error_code, 0)
        expected_assignment = [(b'testtopic_replicated', [6, 7, 8, 9, 0, 1, 2, 3, 4, 5])]
        self.assertEqual(response.member_assignment.partition_assignment,
                         expected_assignment)

    def test_heartbeat_request(self):
        req = protocol.HeartbeatRequest(b'dummygroup', 1, b'testmember')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00-\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\n'  # len(member id)
                    b'testmember'  # member id
            )
        )

    def test_heartbeat_response(self):
        response = protocol.HeartbeatResponse(
            bytearray(
                b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.error_code, 0)

    def test_leave_group_request(self):
        req = protocol.LeaveGroupRequest(b'dummygroup', b'testmember')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00)\x00\r\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id
                b'\x00\n'  # len(member id)
                    b'testmember'  # member id
            )
        )

    def test_leave_group_response(self):
        response = protocol.LeaveGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.error_code, 0)


if __name__ == '__main__':
    unittest2.main()



from __future__ import division

import os
import platform
import pytest
import time
import types
import unittest2
from uuid import uuid4

from pykafka import KafkaClient
from pykafka.common import OffsetType
from pykafka.exceptions import MessageSizeTooLarge, ProducerQueueFullError
from pykafka.partitioners import hashing_partitioner
from pykafka.protocol import Message
from pykafka.test.utils import get_cluster, stop_cluster, retry
from pykafka.common import CompressionType
from pykafka.producer import OwnedBroker


kafka_version = os.environ.get('KAFKA_VERSION', '0.8.0')


class ProducerIntegrationTests(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = b'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.client = KafkaClient(cls.kafka.brokers,
                                 use_greenlets=cls.USE_GEVENT,
                                 broker_version=kafka_version)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def _get_producer(self, **kwargs):
        topic = self.client.topics[self.topic_name]
        return topic.get_producer(use_rdkafka=self.USE_RDKAFKA, **kwargs)

    def _get_consumer(self):
        return self.client.topics[self.topic_name].get_simple_consumer(
            consumer_timeout_ms=1000,
            auto_offset_reset=OffsetType.LATEST,
            reset_offset_on_start=True,
        )

    def test_produce(self):
        # unique bytes, just to be absolutely sure we're not fetching data
        # produced in a previous test
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(sync=True, min_queued_messages=1)
        prod.produce(payload)

        message = consumer.consume()
        assert message.value == payload

    def test_sync_produce_raises(self):
        """Ensure response errors are raised in produce() if sync=True"""
        with self._get_producer(sync=True, min_queued_messages=1) as prod:
            with self.assertRaises(MessageSizeTooLarge):
                prod.produce(10 ** 7 * b" ")

        if not self.USE_RDKAFKA:
            # ensure that a crash on a worker thread still raises exception in sync mode
            p = self._get_producer(sync=True)
            def stub_send_request(self, message_batch, owned_broker):
                1/0
            p._send_request = types.MethodType(stub_send_request, p)
            with self.assertRaises(ZeroDivisionError):
                p.produce(b"test")

    def test_produce_hashing_partitioner(self):
        # unique bytes, just to be absolutely sure we're not fetching data
        # produced in a previous test
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(
            sync=True,
            min_queued_messages=1,
            partitioner=hashing_partitioner)
        prod.produce(payload, partition_key=b"dummy")

        # set a timeout so we don't wait forever if we break producer code
        message = consumer.consume()
        assert message.value == payload

    def test_async_produce(self):
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(min_queued_messages=1, delivery_reports=True)
        prod.produce(payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, payload)
        self.assertIsNone(report[1])

        message = consumer.consume()
        assert message.value == payload

    def test_recover_disconnected(self):
        """Test our retry-loop with a recoverable error"""
        payload = uuid4().bytes
        prod = self._get_producer(min_queued_messages=1, delivery_reports=True)

        for broker in self.client.brokers.values():
            broker._connection.disconnect()

        prod.produce(payload)
        report = prod.get_delivery_report()
        self.assertIsNone(report[1])

    def test_async_produce_context(self):
        """Ensure that the producer works as a context manager"""
        payload = uuid4().bytes

        consumer = self._get_consumer()
        with self._get_producer(min_queued_messages=1) as producer:
            producer.produce(payload)

        message = consumer.consume()
        assert message.value == payload

    def test_async_produce_queue_full(self):
        """Ensure that the producer raises an error when its queue is full"""
        consumer = self._get_consumer()
        with self._get_producer(block_on_queue_full=False,
                                max_queued_messages=1,
                                linger_ms=1000) as producer:
            with self.assertRaises(ProducerQueueFullError):
                while True:
                    producer.produce(uuid4().bytes)
        while consumer.consume() is not None:
            time.sleep(.05)

    def test_async_produce_lingers(self):
        """Ensure that the context manager waits for linger_ms milliseconds"""
        linger = 3
        consumer = self._get_consumer()
        with self._get_producer(linger_ms=linger * 1000) as producer:
            start = time.time()
            producer.produce(uuid4().bytes)
            producer.produce(uuid4().bytes)
        self.assertTrue(int(time.time() - start) >= int(linger))
        consumer.consume()
        consumer.consume()

    def test_async_produce_thread_exception(self):
        """Ensure that an exception on a worker thread is raised to the main thread"""
        consumer = self._get_consumer()
        with self.assertRaises(AttributeError):
            with self._get_producer(min_queued_messages=1) as producer:
                # get some dummy data into the queue that will cause a crash
                # when flushed:
                msg = Message("stuff", partition_id=0)
                del msg.value
                producer._produce(msg)
        while consumer.consume() is not None:
            time.sleep(.05)

    def test_required_acks(self):
        """Test with non-default values for `required_acks`

        See #278 for a related bug.  Here, we only test that no exceptions
        occur (hence `sync=True`, which would surface most exceptions)
        """
        kwargs = dict(linger_ms=1, sync=True, required_acks=0)
        prod = self._get_producer(**kwargs)
        prod.produce(uuid4().bytes)

        kwargs["required_acks"] = -1
        prod = self._get_producer(**kwargs)
        prod.produce(uuid4().bytes)

    def test_null_payloads(self):
        """Test that None is accepted as a null payload"""
        consumer = self._get_consumer()
        prod = self._get_producer(sync=True, min_queued_messages=1)
        prod.produce(None)
        self.assertIsNone(consumer.consume().value)
        prod.produce(None, partition_key=b"whatever")
        self.assertIsNone(consumer.consume().value)
        prod.produce(b"")  # empty string should be distinguished from None
        self.assertEqual(b"", consumer.consume().value)

    def test_owned_broker_flush_message_larger_then_max_request_size(self):
        """Test that producer batches messages into the batches no larger than
        `max_request_size`
        """
        large_payload = b''.join([uuid4().bytes for i in range(50000)])

        producer = self._get_producer(
            auto_start=False
        )
        # setup producer but do not start actually sending messages
        partition = producer._topic.partitions[0]
        owned_broker = OwnedBroker(producer, partition.leader, auto_start=False)

        delivery_report_queue = producer._cluster.handler.Queue()
        msg = Message(
            large_payload,
            partition_id=0,
            delivery_report_q=delivery_report_queue
        )

        owned_broker.enqueue(msg)

        max_request_size = 1000
        assert max_request_size < len(msg)
        owned_broker.flush(0, max_request_size)
        q_msg, exc = delivery_report_queue.get()
        assert q_msg is msg
        assert isinstance(exc, MessageSizeTooLarge)

    def test_owned_broker_flush_batching_by_max_request_size(self):
        """Test that producer batches messages into the batches no larger than
        `max_request_size`
        """
        large_payload = b''.join([uuid4().bytes for i in range(5000)])

        producer = self._get_producer(
            auto_start=False
        )
        # setup producer but do not start actually sending messages
        partition = producer._topic.partitions[0]
        owned_broker = OwnedBroker(producer, partition.leader, auto_start=False)

        for i in range(100):
            msg = Message(large_payload, partition_id=0)
            owned_broker.enqueue(msg)

        batch = owned_broker.flush(0, producer._max_request_size)
        assert len(batch) < 100
        assert sum([len(m.value) for m in batch]) < producer._max_request_size

        # iterate through the rest of the batches and test the same invariant
        while batch:
            batch = owned_broker.flush(0, producer._max_request_size)
            assert len(batch) < 100
            assert sum([len(m.value) for m in batch]) < producer._max_request_size

    def test_async_produce_compression_large_message(self):
        # TODO: make payload size bigger once pypy snappy compression issue is
        # fixed
        large_payload = b''.join([uuid4().bytes for i in range(5)])
        consumer = self._get_consumer()

        prod = self._get_producer(
            compression=CompressionType.SNAPPY,
            delivery_reports=True
        )
        prod.produce(large_payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, large_payload)
        self.assertIsNone(report[1])

        message = consumer.consume()
        assert message.value == large_payload

        for i in range(10):
            prod.produce(large_payload)

        # use retry logic to loop over delivery reports and ensure we can
        # produce a group of large messages
        reports = []

        def ensure_all_messages_produced():
            report = prod.get_delivery_report()
            reports.append(report)
            assert len(reports) == 10
        retry(ensure_all_messages_produced, retry_time=30, wait_between_tries=0.5)

        for report in reports:
            self.assertEqual(report[0].value, large_payload)
            self.assertIsNone(report[1])

        # cleanup and consumer all messages
        msgs = []

        def ensure_all_messages_consumed():
            msg = consumer.consume()
            if msg:
                msgs.append(msg)
            assert len(msgs) == 10
        retry(ensure_all_messages_consumed, retry_time=15)

    def test_async_produce_large_message(self):

        consumer = self._get_consumer()
        large_payload = b''.join([uuid4().bytes for i in range(50000)])
        assert len(large_payload) / 1024 / 1024 < 1.0

        prod = self._get_producer(delivery_reports=True, linger_ms=1000)
        prod.produce(large_payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, large_payload)
        self.assertIsNone(report[1])

        message = consumer.consume()
        assert message.value == large_payload

        for i in range(10):
            prod.produce(large_payload)

        # use retry logic to loop over delivery reports and ensure we can
        # produce a group of large messages
        reports = []

        def ensure_all_messages_produced():
            report = prod.get_delivery_report()
            reports.append(report)
            assert len(reports) == 10
        retry(ensure_all_messages_produced, retry_time=30, wait_between_tries=0.5)

        for report in reports:
            self.assertEqual(report[0].value, large_payload)
            self.assertIsNone(report[1])

        # cleanup and consumer all messages
        msgs = []

        def ensure_all_messages_consumed():
            msg = consumer.consume()
            if msg:
                msgs.append(msg)
            assert len(msgs) == 10
        retry(ensure_all_messages_consumed, retry_time=15)


@pytest.mark.skipif(platform.python_implementation() == "PyPy",
                    reason="Unresolved crashes")
class TestGEventProducer(ProducerIntegrationTests):
    USE_GEVENT = True


if __name__ == "__main__":
    unittest2.main()



import platform

import pytest

from tests.pykafka import test_producer


@pytest.mark.skipif(platform.python_implementation() == "PyPy",
                    reason="We pass PyObject pointers as msg_opaques for "
                           "delivery callbacks, which is unsafe on PyPy.")
class TestRdKafkaProducer(test_producer.ProducerIntegrationTests):
    USE_RDKAFKA = True

    @pytest.mark.xfail
    def test_async_produce_lingers(self):
        """Async produce may not linger

        There seems to be a difference in interpretation between pykafka's
        `linger_ms` and librdkafka's `queue.buffering.max.ms`: in librdkafka,
        it is the longest a message may stick around, but it will be shipped
        sooner if that is expedient.  We cannot, therefore, comply with this
        particular test, but nor do I see practical problems with that.
        """
        super(TestRdKafkaProducer, self).test_async_produce_lingers()

    @pytest.mark.xfail
    def test_recover_disconnected(self):
        """Test fails because we don't use the pykafka.Brokers after start

        (But it won't matter either, because the librdkafka producer does its
        own connection recovery.)
        """
        super(TestRdKafkaProducer, self).test_recover_disconnected()
